import re
import json

from decimal import Decimal

from ki_errors import EOF, ParseError

class Regex:

    def __init__(self, rx):
        self.rx = re.compile(rx)
        self.match = None

    def matches(self, s):
        self.match = self.rx.match(s)
        return self.match is not None

class KiLexer:

    def __init__(self, defs_file, input_file):
        self.line = 1
        self.column = 1
        with open(defs_file, 'r') as defs_fobj:
            self.load_defs(json.load(defs_fobj))
        self.input_file = input_file

    def load_defs(self, defs):
        self.types = defs['types']
        self.statements = defs['statements']
        self.statement_functions = defs['statement functions']
        self.keywords = defs['keywords']
        self.declarators = defs['declarators']
        self.bind_types = defs['bind types']
        self.math_operators = defs['math operators']
        self.boolean_operators = defs['boolean operators']
        self.assignment_operators = defs['assignment operators']
        self.operators = {}
        self.operators.update(self.math_operators)
        self.operators.update(self.boolean_operators)
        self.operators.update(self.assignment_operators)
        self.symbols = defs['symbols']
        self.rune_literal_rx = Regex(defs['rune literal'])
        self.string_literal_rx = Regex(defs['string literal'])
        self.identifier_rx = Regex(defs['identifier'])
        self.call_rx = Regex(defs['call'])
        self.type_literal_rx = Regex(defs['type literal'])
        self.integer_literal_rx = Regex(defs['integer literal'])
        self.hex_literal_rx = Regex(defs['hex literal'])
        self.binary_literal_rx = Regex(defs['binary literal'])
        self.decimal_literals = [
            (x, Regex(y)) for x, y in defs['decimal literals']
        ]

    def get_token_from_buffer(self):
        buf = self.buf[self.cursor:]
        if self.rune_literal_rx.matches(buf):
            match = self.rune_literal_rx.match
            return match.group(), {
                'type': 'rune literal',
                'value_type': 'rune',
                'value': match.group(1)
            }
        if self.string_literal_rx.matches(buf):
            match = self.string_literal_rx.match
            return match.group(0), {
                'type': 'string literal',
                'value_type': 'string',
                'value': match.group(0)
            }
        if self.call_rx.matches(buf):
            match = self.call_rx.match
            d = {'type': 'call'}
            pointer_type = match.group(1)
            if not pointer_type:
                d['pointer_type'] = None
            elif pointer_type == '*':
                d['pointer_type'] = 'unique'
            elif pointer_type == '@':
                d['pointer_type'] = 'shared'
            elif pointer_type == '$':
                d['pointer_type'] = 'weak'
            elif pointer_type == '&':
                d['pointer_type'] = 'reference'
            else:
                raise ParseError(self.input_file, self.line, self.column,
                                 'Invalid pointer type {}'.format(pointer_type))
            d['value'] = match.group(2)
            return match.group(), d
        if self.type_literal_rx.matches(buf):
            match = self.type_literal_rx.match
            d = {'type': 'type literal', 'value_type': 'type'}
            pointer_type = match.group(1)
            if not pointer_type:
                d['pointer_type'] = None
            elif pointer_type == '*':
                d['pointer_type'] = 'unique'
            elif pointer_type == '@':
                d['pointer_type'] = 'shared'
            elif pointer_type == '$':
                d['pointer_type'] = 'weak'
            elif pointer_type == '&':
                d['pointer_type'] = 'reference'
            else:
                raise ParseError(self.input_file, self.line, self.column,
                                 'Invalid pointer type {}'.format(pointer_type))
            d['value'] = match.group(2)
            mutability = match.group(3)
            if mutability == '!':
                d['mutable'] = True
            elif not mutability:
                d['mutable'] = False
            else:
                msg = 'Invalid mutability value {}'.format(mutability)
                raise ParseError(self.input_file, self.line, self.column, msg)
            return match.group(), d
        if self.identifier_rx.matches(buf):
            match = self.identifier_rx.match
            identifier = match.group()
            d = {}
            pointer_type = match.group(1)
            if not pointer_type:
                pointer_type = None
            elif pointer_type == '*':
                pointer_type = 'unique'
            elif pointer_type == '@':
                pointer_type = 'shared'
            elif pointer_type == '$':
                pointer_type = 'weak'
            elif pointer_type == '&':
                pointer_type = 'reference'
            else:
                raise ParseError(self.input_file, self.line, self.column,
                                 'Invalid pointer type {}'.format(pointer_type))
            mutable = match.group(3)
            if mutable == '!':
                mutable = True
            elif mutable is None:
                mutable = False
            else:
                msg = 'Invalid mutability value {}'.format(mutable)
                raise ParseError(self.input_file, self.line, self.column, msg)
            if identifier in self.types:
                category = 'type'
            elif identifier in self.statements:
                if identifier in self.statement_functions:
                    msg = 'Statement "{}" must be used as a function'
                    raise ParseError(self.input_file,
                                     self.line,
                                     self.column,
                                     msg.format(identifier))
                category = 'statement'
            elif identifier in self.keywords:
                category = 'keyword'
            elif identifier in self.declarators:
                category = 'declarator'
            elif identifier in self.bind_types:
                category = 'bind type'
            else:
                category = 'identifier'
            if category in ('type', 'identifier'):
                return identifier, {
                    'pointer_type': pointer_type,
                    'mutable': mutable,
                    'value': identifier,
                    'type': 'identifier',
                    'category': category,
                }
            return identifier, {
                'value': identifier,
                'type': 'identifier',
                'category': category,
            }
        if self.hex_literal_rx.matches(buf):
            match = self.hex_literal_rx.match
            value = Decimal(int(match.group(1), 16))
            value_type = match.group(2)
            if value_type == 'u':
                value_type = 'uint'
            elif value_type == 's' or value_type is None:
                value_type = 'int'
            return match.group(), {
                'type': 'hex literal',
                'value_type': value_type,
                'value': value
            }
        if self.binary_literal_rx.matches(buf):
            match = self.binary_literal_rx.match
            value = Decimal(int(match.group(1), 2))
            value_type = match.group(2)
            if value_type == 'u':
                value_type = 'uint'
            elif value_type == 's' or value_type is None:
                value_type = 'int'
            return match.group(), {
                'type': 'binary literal',
                'value_type': value_type,
                'value': value
            }
        for rx_name, rx in self.decimal_literals:
            if rx.matches(buf):
                match = rx.match
                if match.group().endswith('f32'):
                    width = '32'
                    lit = match.group()[:-3]
                elif match.group().endswith('f64'):
                    width = '64'
                    lit = match.group()[:-3]
                elif match.group().endswith('d'):
                    width = 'arbitrary'
                    lit = match.group()[:-1]
                else:
                    width = 'arbitrary'
                    lit = match.group()
                return match.group(), {
                    'type': 'decimal literal',
                    'value_type': width,
                    'value': Decimal(lit)
                }
        if self.integer_literal_rx.matches(buf):
            match = self.integer_literal_rx.match
            if match.group(2):
                value = Decimal(match.group(2))
                value_type = match.group(3)
            else:
                value = Decimal(match.group(5))
                value_type = match.group(6)
            if value_type == 'u':
                value_type = 'uint'
            elif value_type == 's' or value_type is None:
                value_type = 'int'
            return match.group(), {
                'type': 'integer literal',
                'value_type': value_type,
                'value': value
            }
        for op in sorted(self.operators, reverse=True, key=lambda x: len(x)):
            if buf.startswith(op):
                if op in self.math_operators:
                    d = {
                        'type': 'math operator',
                        'value': op
                    }
                    d.update(self.math_operators[op])
                    return op, d
                elif op in self.boolean_operators:
                    d = {
                        'type': 'boolean operator',
                        'value': op
                    }
                    d.update(self.boolean_operators[op])
                    return op, d
                elif op in self.assignment_operators:
                    d = {
                        'type': 'assignment operator',
                        'value': op,
                    }
                    d.update(self.assignment_operators[op])
                    return op, d
                else:
                    raise Exception('Unknown operator type for {}'.format(op))
        for symbol in self.symbols:
            if buf.startswith(symbol):
                if symbol == '#':
                    comment_end = buf.find('\n')
                    if comment_end == -1:
                        comment_end = len(buf)
                    return buf[:comment_end], {
                        'value': buf[:comment_end],
                        'type': 'comment'
                    }
                return symbol, {
                    'value': symbol,
                    'type': 'symbol'
                }

    def advance_counters(self):
        while self.buf[self.cursor].isspace():
            if self.buf[self.cursor] == '\n':
                self.line += 1
                self.column = 1
                self.cursor += 1
            else:
                self.cursor += 1
                self.column += 1
            if self.cursor == len(self.buf):
                raise EOF()

    @property
    def tokens(self):
        with open(self.input_file, 'r', encoding='utf-8') as fobj:
            self.buf = fobj.read()
        self.line = 1
        self.column = 1
        self.cursor = 0
        while self.cursor < len(self.buf):
            try:
                self.advance_counters()
            except EOF:
                break
            result = self.get_token_from_buffer()
            if result:
                match, token = result
                token['line'] = self.line
                token['column'] = self.column
                yield token
                self.cursor += len(match)
                self.column += len(match)
                try:
                    self.advance_counters()
                except EOF:
                    break
            else:
                sample = repr(self.buf[self.cursor:self.cursor+20])
                raise ParseError(self.input_file, self.line, self.column,
                     'Cannot handle input: [{}]'.format(sample)
                )
